
class XResultSet
{
    public:
	int size;
    	Xapian::docid * data;

    XResultSet() { size=0; data=NULL; }
    ~XResultSet() { if (size>0) { free(data); } }

    void add(Xapian::docid did)
    {
        if(data==NULL)
        {
            data=(Xapian::docid *)malloc(sizeof(Xapian::docid));
        }
        else
        {
            data=(Xapian::docid *)realloc(data,(size+1)*sizeof(Xapian::docid));
        }
        data[size]=did;
        size++;
    }
};
    
class XQuerySet
{
    	public:
		int size;
    		char ** data;
    		bool is_and;

    	XQuerySet() { size=0; data=NULL; }
    	~XQuerySet() { if (size>0) { for(int i=0;i<size;i++) { free(data[i]); } free(data); } }

	char * clean(const char *s)
  	{
		int i=0;
        	char * s2 = (char *)malloc(sizeof(char)*(strlen(s)+1));
          	while((s[i]<=' ') && (s[i]>0))
          	{
			i++;
		}
		strcpy(s2, s+i);
          	
		while((strlen(s2)>0) && (s2[strlen(s2)-1]<=' '))
          	{
              		s2[strlen(s2)-1]=0;
          	}
          	for(int i=0;i<strlen(s2);i++)
          	{
              		s2[i]=tolower(s2[i]);
          	}
          	return s2;
  	}

    	void add(const char * type,const char * s)
    	{
        	if(s==NULL) return;
        	if(type==NULL) return;
        
		char * s2=clean(s);
        	if(strlen(s2)<1) { free(s2); return; }

        	char * blank=strstr(s2," ");
        	while(blank!=NULL)
        	{
            		blank[0]=0;
            		add(type,blank+1);
        	}

        	char *t2=clean(type);
        	if(strlen(t2)>0)
        	{
            		char *s3=(char *)malloc(sizeof(char)*(strlen(s2)+strlen(t2)+1));
            		sprintf(s3,"%s:%s",t2,s2);
            		free(s2);
            		s2=s3;
        	}
        	free(t2);
        
        	if(data==NULL)
        	{
            		data=(char **)malloc(sizeof(char*));
        	}
        	else
        	{
            		data=(char **)realloc(data,(size+1)*sizeof(char*));
        	}
        	data[size]=s2;
        	size++;
    	}

    	char * get_query()
    	{
        	long n=0;
        	int i,j;

        	for(i=0;i<size;i++)
        	{
            		n=n+strlen(data[i])+5;
        	}
        	char * s = (char *)malloc(sizeof(char)*(n+1));
        	s[0]=0;
        	n=0;
        	for(i=0;i<size;i++)
        	{
            		if(i>0)
            		{
                		if(this->is_and)
                		{                
                    			strcpy(s+n," AND ");
                    			n+=5;
                		}
                		else
                		{
                    			strcpy(s+n," OR ");
                    			n+=4;
                		}
            		}
            		strcpy(s+n,data[i]);
            		n+=strlen(data[i]);
        	}
        	return s;
    	}
};

class XHeaderTerm
{
	public:
		int size,partial,full;
		char ** data;
  
        XHeaderTerm(int p, int f) { partial=p; full=f; size=0; data=NULL; }
        ~XHeaderTerm() { if (size>0) { for(int i=0;i<size;i++) { free(data[i]); } free(data); } }

	char * strip(const char *s)
	{
		int i=0;
		char * s2 = (char *)malloc(sizeof(char)*(strlen(s)+1));
		while(i<strlen(s))
		{
			if((s[i]=='"') || (s[i]=='<') || (s[i]=='>'))
			{
				s2[i]=' ';
			}
			else
			{
				s2[i]=s[i];
			}
			i++;
		}
		s2[i]=0;
		return s2;
	}

	char * clean(const char *s)
        {
        	int i=0;
                char * s2 = (char *)malloc(sizeof(char)*(strlen(s)+1));
                while((s[i]<=' ') && (s[i]>0))
                {
                	i++;
                }
                strcpy(s2, s+i);
  
                while((strlen(s2)>0) && (s2[strlen(s2)-1]<=' '))
                {
                	s2[strlen(s2)-1]=0;
                }
                for(int i=0;i<strlen(s2);i++)
                {
                        s2[i]=tolower(s2[i]);
                }
                return s2;
        }

	void add(const char * s)
        {
        	if(s==NULL) return;
  
                char * s2=strip(s);
		char * s3=clean(s2);
		free(s2);
		
                if(strlen(s3)<1) { free(s3); return; }
  
                char * blank=strstr(s3," ");
                while(blank!=NULL)
                {
                	blank[0]=0;
                        add(blank+1);
			blank=strstr(blank+1," ");
                }

		int i,j;

		char *stem = (char *)malloc(sizeof(char)*(full+1)); 
		for(i=0;i<strlen(s3);i++)
		{
			for(j=partial;j<strlen(s3)-i;j++)
			{
				strncpy(stem,s3+i,j);
				stem[j]=0;
				add_stem(stem);
			}
		}
		add_stem(s3);
	}
	
	void add_stem(const char *s)
	{
		char * s2 = (char *)malloc(sizeof(char)*(strlen(s)+1));
		strcpy(s2,s);
				
                if(data==NULL)
                {
                	data=(char **)malloc(sizeof(char*));
                }
                else
                {
			bool existing=false;
			int i=0;
			while((!existing) && (i<size))
                	{
				if(strcmp(data[i],s2)==0)
				{
					existing=true;
				}
			}
			if(existing) return;
                	data=(char **)realloc(data,(size+1)*sizeof(char*));
                }
                data[size]=s2;
                size++;
	}
};

static int fts_backend_xapian_unset_box(struct xapian_fts_backend *backend)
{
	backend->box = NULL;
	backend->db[0] = 0;
	backend->path[0]=0;
	return 0;
}

static int fts_backend_xapian_set_box(struct xapian_fts_backend *backend, struct mailbox *box)
{
	fts_backend_xapian_unset_box(backend);

	if (box == NULL)
	{
                return 0;
	}

	i_warning("fts_xapian : Set BOX %s",box->name);

	const char * mb;
	if (mailbox_get_path_to(box, MAILBOX_LIST_PATH_TYPE_INDEX, &mb) <= 0)
                i_unreached(); /* fts already checked this */
	sprintf(backend->path,"%s/%s",mb,XAPIAN_FILE_PREFIX);
	i_warning("fts_xapian: Set BOX PATH : %s", backend->path);
/*
        struct mailbox_metadata metadata;
        if (mailbox_get_metadata(box, MAILBOX_METADATA_GUID, &metadata) < 0) 
	{
                i_error("fts_xapian: Couldn't get mailbox %s GUID: %s", box->vname, mailbox_get_last_internal_error(box, NULL));
                return -1;
        }
        memcpy(guid_r, metadata.guid, GUID_128_SIZE);
*/
	fts_mailbox_get_guid(box, &mb );
	strcpy(backend->guid,mb);
	sprintf(backend->db,"%s/db_%s",backend->path,backend->guid);

	i_warning("fts_xapian: Set BOX DB : %s", backend->db);

	Xapian::WritableDatabase db(backend->db,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK);
    
	backend->box = box;
	return 0;
}

XResultSet * fts_backend_xapian_query(char * db_path, XQuerySet * query, long limit=0)
{
    XResultSet * set= new XResultSet();
    
    try
    {
        Xapian::Database db(db_path);
        Xapian::Enquire enquire(db);

        Xapian::QueryParser * qp = new Xapian::QueryParser;
        qp->add_prefix("uid", "Q");
        qp->add_prefix("subject", "S");
        qp->add_prefix("from", "A");
        qp->add_prefix("to", "XTO");
        qp->add_prefix("cc", "XCC");
        qp->add_prefix("bcc", "XBCC");
        qp->add_prefix("", "XBDY");
        qp->set_database(db);
    
        char *query_string=query->get_query();
        Xapian::Query q = qp->parse_query(query_string);
        free(query_string);

        enquire.set_query(q);

        long offset=0;
        long pagesize=100; if(limit>0) { pagesize=std::min(pagesize,limit); }
        Xapian::MSet m = enquire.get_mset(0, pagesize);
        while(m.size()>0)
        {
            Xapian::MSetIterator i = m.begin();
	        while (i != m.end()) 
            {
                Xapian::Document doc = i.get_document();
	            set->add(doc.get_docid());
	            i++;
            }
            offset+=pagesize;
            m = enquire.get_mset(offset, pagesize);
        }
    }
    catch(Xapian::Error e)
    {
        i_error(e.get_msg().c_str());
    }
    return set;
}

bool fts_backend_xapian_index_hdr(char * db_path,uint uid, char* field, const char* data,int p, int f)
{
	try
	{
		XQuerySet xq;
        	char u[1000]; sprintf(u,"%i",uid);
        	xq.add("uid",u);
        	XResultSet *result=fts_backend_xapian_query(db_path,&xq);

		Xapian::WritableDatabase db(db_path,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK);

		Xapian::docid docid;
		Xapian::Document doc;
		if(result->size<1)
        	{
			doc.add_value(1,u);
			sprintf(u,"New Document for UID=%i",uid);
        	        i_warning(u);
			docid=db.add_document(doc);
        	}
		else
		{
			docid=result->data[0];
			doc = db.get_document(docid);
		}
		free(result);

		XHeaderTerm xhs(p,f);
		xhs.add(data);
	
	        int i=0;
	        char * f2 = (char *)malloc(sizeof(char)*(strlen(field)+1));
		while((field[i]<=' ') && (field[i]>0))
		{
			i++;
		}
	        strcpy(f2, field+i);
	  
		while((strlen(f2)>0) && (f2[strlen(f2)-1]<=' '))
		{
			f2[strlen(f2)-1]=0;
		}
		for(int i=0;i<strlen(f2);i++)
		{
			f2[i]=tolower(f2[i]);
		}
		char h[10]; h[0]=0;
		if(strcmp(f2,"uid")==0) { strcpy(h,"Q"); }
		else if(strcmp(f2,"from")==0) { strcpy (h,"A"); }
		else if(strcmp(f2,"to")==0) { strcpy(h,"XTO"); }
		else if(strcmp(f2,"cc")==0) { strcpy(h,"XCC"); }
		else if(strcmp(f2,"bcc")==0) { strcpy(h,"XBCC"); }
		else 
		{
			sprintf(u,"Unknown field %s",f2);
			i_warning(u);
			return false;
		}
	
		for(int i=0;i<xhs.size;i++)
		{
			sprintf(u,"%s%s",h,xhs.data[i]);		
			
			doc.add_term(u);
		}

		db.replace_document(docid,doc);
	    	return true;
	}
	catch(Xapian::Error e)
	{
		i_error(e.get_msg().c_str());
	}
	return false;
}

bool fts_backend_xapian_index_text(char * db_path,uint uid, char * field, const char * data)
{
	try
        {
        	XQuerySet xq;
                char u[1000]; sprintf(u,"%i",uid);
                xq.add("uid",u);
                XResultSet * result=fts_backend_xapian_query(db_path,&xq);
  
                Xapian::WritableDatabase db(db_path,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK);
  
                Xapian::docid docid;
                Xapian::Document doc;
                if(result->size<1)
                {
                	doc.add_value(1,u);
                        sprintf(u,"New Document for UID=%i",uid);
                        i_warning(u);
                        docid=db.add_document(doc);
                }
                else
                {
                        docid=result->data[0];
                        doc = db.get_document(docid);
                }
		free(result);

		Xapian::TermGenerator termgenerator;
		termgenerator.set_stemmer(Xapian::Stem("en"));
		termgenerator.set_document(doc);
	
		std::string h;
                if(strcmp(field,"subject")==0) 
		{
			h="S";
		}
		else
		{
			h="XBDY";
		}
		std::string d(data);
		termgenerator.index_text(d, 1, h);
		
                db.replace_document(docid,doc);
                return true;
          }
          catch(Xapian::Error e)
          {
                  i_error(e.get_msg().c_str());
          }
          return false;
}

